#!/usr/bin/perl -w -T
########################################################
#
# This script is intended to provide a means for
# detecting changes made to files, via a regular
# comparison of MD5 hashes to an established
# "baseline".
# In this respect, it is designed as a portable clone
# of aide software.
#
# This script requires perl ,and some others perl modules
# which come in standard installation
#
###############################################################################
#    Copyright (C) 2002 by Eric Gerbier
#    Bug reports to: gerbier@users.sourceforge.net
#    $Id: afick.pl,v 1.33 2003/11/13 09:34:27 gerbier Exp $
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
###############################################################################

#######################################################
use strict;

use Digest::MD5;    # for md5 checksum

use SDBM_File;                                  # database
use FileHandle;                                 # for database open mode
use Getopt::Long;    # arg analysis

#use Pod::Usage;
use File::Glob ':glob';                         # for jokers
use File::Basename;                             # for delete

################# global variables #############################

use vars qw ($Dbm);

$Dbm = 'SDBM_File';                             # database type

use vars
  qw (%Finded %Hashfile %Newfiles %Dangling %Toscan %Onlydir %Onlythis %Oldval %Newval);
%Finded   = ();    # list of files scanned
%Hashfile = ();    # dbm hash-table
%Newfiles = ();    # list of new files
%Dangling = ();    # dangling links
%Toscan   = ();    # list to scan
%Onlydir  = ();    # only this and content, without sub-dir
%Onlythis = ();    # only this inode
%Oldval   = ();    # store old values
%Newval   = ();    # store new changes values

use vars qw ( $Nbmod  $Update $File);
$Nbmod  = 0;       # number of modifications
$Update = 0;       # update/compare mode ?
$File   = 0;       # file mode ?

# directives status
# not initialized to be able to detect negative command lines
use vars
  qw ($Verbose $Warn_dead_symlinks $Report_full_newdel $Warn_missing_file $Report_url);

#$Verbose            ;           # verbose mode ?
#$Warn_dead_symlinks = '';           # warn_dead_symlink
#$Report_full_newdel     = '';	   # report new directories contents
#$Warn_missing_file   = '';
$Report_url = "STDOUT";    # report_url

################ end of global var #############################

my @listerep;              # list of rep to scan
my $database;              # database name

# command line
my $configfile;            # config file name
my $help;
my $init;
my $compare;
my $list;
my $print;
my $Version;
my $update;

my $database_ext = ".pag";    # database extension
my $control_ext  = ".ctr";    # database extension

my $version = "1.4-0";

my $default_conf;

#######################################################
# get md5 from database
sub get_md5 ($) {
	my $name = shift (@_);    # file name

	my $hash = $Hashfile{$name} || $Newval{$name} || "";
	my @tab      = split ( /\|/, $hash );
	my $checksum = $tab[0];

	return $checksum;
}
#######################################################
# get file_type from database
sub file_type ($) {
	my $name = shift (@_);    # file name

	# see sys/stat.h
	my $S_IFIFO  = 0010000;
	my $S_IFCHR  = 0020000;
	my $S_IFDIR  = 0040000;
	my $S_IFBLK  = 0060000;
	my $S_IFREG  = 0100000;
	my $S_IFLNK  = 0120000;
	my $S_IFSOCK = 0140000;

	# $Hashfile for changed file
	# Newval in compare mode for new files
	my $hash = $Hashfile{$name} || $Newval{$name};
	my @tab  = split ( /\|/, $hash );
	my $perm = $tab[3];

	if ( ( $perm & $S_IFSOCK ) == $S_IFSOCK ) {
		return "socket";
	}
	elsif ( ( $perm & $S_IFLNK ) == $S_IFLNK ) {
		return "symbolic link";
	}
	elsif ( ( $perm & $S_IFREG ) == $S_IFREG ) {
		return "file";
	}
	elsif ( ( $perm & $S_IFBLK ) == $S_IFBLK ) {
		return "block device";
	}
	elsif ( ( $perm & $S_IFDIR ) == $S_IFDIR ) {
		return "directory";
	}
	elsif ( ( $perm & $S_IFCHR ) == $S_IFCHR ) {
		return "character device";
	}
	elsif ( ( $perm & $S_IFIFO ) == $S_IFIFO ) {
		return "fifo";
	}
	else {
		return "unknown type";
	}

}
#######################################################
# used to send all "normal" messages
sub report($) {
	no strict 'refs';
	print $Report_url @_ if ($Report_url);
}
#######################################################
# for debug purpose
sub debug($) {
	print "DEBUG : @_" if ($Verbose);
}
#######################################################
# remove_trailing blanks
sub remove_trailing_blanks ($) {
	my $val = shift (@_);
	$$val =~ s/^\s+//;
	$$val =~ s/\s+$//;
}
#######################################################
# test_double
# to detect bad configuration
sub test_double($) {
	my $elem = shift (@_);
	return exists $Toscan{$elem};
}
#######################################################
# expand_line
# expand config line : treat jokers, and split into <file attribute flagdir>
sub trait_line ($) {
	my $line = shift (@_);

	my $name;         # file name
	my $attribute;    # what to do on $name

	# for file with blank caracters
	if ( $line =~ m/^"(.*)"(.*)/ ) {
		$name      = $1;
		$attribute = $2;
	}
	else {
		( $name, $attribute ) = split ( " ", $line, 2 );
	}

	# compatibilite with old syntaxe
	# default is "all"
	$attribute = "all" if ( !$attribute );

	# debug( "$name -> $attribute");

	my @tab;
	my $flagdir;

	# expand jokers if exists
	if ( $name =~ m/[\*\?]/ ) {
		@tab = glob($name);
		debug("expand $name in @tab\n");
	}
	elsif ( $name eq "/" ) {
		$flagdir = 1;
		push ( @tab, $name );
	}
	elsif ( $name =~ m?/$? ) {

		# remove trailing /
		$name =~ s?/$??;
		$flagdir = 1;
		push ( @tab, $name );
	}
	else {
		$flagdir = 0;
		push ( @tab, $name );
	}
	push ( @tab, $attribute, $flagdir );
	return @tab;
}
#######################################################
#resolv_alias
# expand and resolv action pattern, for alias definitions and files config
sub resolv_alias($$@) {
	my $nline = shift (@_);    # line number in config file
	my $val   = shift (@_);    # value
	my %alias = @_;            # alias tab

	my $realval;

	$val =~ s/\s*//g;          # remove all blanks
	$val =~ m/^([^+-]*)/;      # first motif before any +/-
	if ( exists $alias{$1} ) {
		$realval = $alias{$1};
	}
	else {

		# error : pattern not found
		warn
"WARNING : error in config file (line $nline) : \"$1\" can not be resolved\n";
	}

	# then other flags
	while ( $val =~ m/-([^+-])/g ) {
		$realval =~ s/$1//;
	}

	while ( $val =~ m/\+([^+-])/g ) {
		$realval .= "$1" if ( $realval !~ m/$1/ );
	}

	if ( !$realval ) {
		$realval = $alias{all};
		warn
		  "WARNING : set to default pattern \"all\" in config line line $.\n";
	}
	return $realval;

}
#######################################################
# decode
# used to read directives : yes/1/true or no/0/false

sub decode ($$$) {
	my $val = lc( shift (@_) );
	my $line   = shift (@_);    # line
	my $nuline = shift (@_);    # number of line

	if ( ( $val eq "yes" ) or ( $val eq "true" ) or ( $val eq "1" ) ) {
		return 1;
	}
	elsif ( ( $val eq "no" ) or ( $val eq "false" ) or ( $val eq "0" ) ) {
		return 0;
	}
	else {
		warn "WARNING : bad value $val in config line line $nuline : $line\n";
		return 0;
	}
}
#######################################################
# read_config :
# read the config file to build a list of directory to scan

sub read_config($) {
	my $config = shift (@_);    # config file name

	my @liste_toscan;

	# alias table
	# contains all "good" values
	my %alias = (
		a   => "a",              # atime
		b   => "b",              # blocks
		c   => "c",              # ctime
		d   => "d",              # device
		g   => "g",              # gid
		h   => "h",              # md5 hash
		i   => "i",              # inode
		m   => "m",              # mtime
		md5 => "h",              # md5 hash (alias)
		n   => "n",              # number of links
		p   => "p",              # permissions
		u   => "u",              # uid
		s   => "s",              # size
		all => "pdinugsbmch",    # all except a
		R   => "pdinugsmch",     # aide compatibility
		L   => "pdinug",         # aide compatibility
		E   => "",               # aide compatibility
	);

	debug("-------- begin of config file -------------\n");

	open( INIT, "< $config" )
	  or die "abort : can not open config file $config : $!\n";

	while ( my $line = <INIT> ) {
		chomp $line;

		$line =~ s/#.*//;        # skip comments
		remove_trailing_blanks( \$line );
		$line =~ s?\\?/?g;       # replace \ by / (for windows)

		next unless length($line);

		#debug( "config line = $line\n");

		if ( $line =~ m/\@\@/ ) {

			# skip macros lines
			debug("skipped macro in line $. : $line\n");
			next;
		}
		elsif ( $line =~ m/^=/ ) {

			# begin with a = : a directory to scan without recurse
			$line =~ s/^=\s*//;

			my @tab     = trait_line($line);
			my $flagdir = pop (@tab);
			my $masqb   = pop (@tab);
			my $masq    = resolv_alias( $., $masqb, %alias );

			foreach my $elem (@tab) {
				if ( test_double($elem) ) {
					warn
"WARNING : skipped config file $line (line $.) : $elem already in config file\n";
				}
				elsif ( -d $elem ) {
					push ( @liste_toscan, $elem );
					$Toscan{$elem} = $masq;
					if ($flagdir) {
						$Onlydir{$elem} = 1;
						debug("toscan without sub-dir : $elem masq $masq\n");
					}
					else {
						$Onlythis{$elem} = 1;
						debug("toscan only this inode : $elem masq $masq\n");
					}
				}
				elsif ( -e $elem ) {
					warn
"WARNING : skipped config file $elem (line $.) : $elem not a directory\n";
				}
				else {
					warn
"WARNING : skipped config file $line (line $.) : $elem does not exists\n";
				}
			}    # foreach my $elem
		}
		elsif ( $line =~ m/:=/ ) {

			# a configuration line
			my ( $key, $val ) = split ( /:=/, $line, 2 );
			remove_trailing_blanks( \$val );

			if ( $key =~ m/^database/ ) {
				debug("found database : $val\n");
				$database = $val;
			}
			elsif ( $key =~ m/^verbose/ ) {
				if ( !defined $Verbose ) {
					$Verbose = decode( $val, $line, $. );
				}
				else {
					debug(
"ignore $line (line $.) : overload by command line option\n"
					);
				}
			}
			elsif ( $key =~ m/^warn_dead_symlinks/ ) {
				if ( !defined $Warn_dead_symlinks ) {
					$Warn_dead_symlinks = decode( $val, $line, $. );
				}
				else {
					debug(
"ignore $line (line $.) : overload by command line option\n"
					);
				}
			}
			elsif ( $key =~ m/^report_full_newdel/ ) {
				if ( !defined $Report_full_newdel ) {
					$Report_full_newdel = decode( $val, $line, $. );
				}
				else {
					debug(
"ignore $line (line $.) : overload by command line option\n"
					);
				}
			}
			elsif ( $key =~ m/^warn_missing_file/ ) {
				if ( !defined $Warn_missing_file ) {
					$Warn_missing_file = decode( $val, $line, $. );
				}
				else {
					debug(
"ignore $line (line $.) : overload by command line option\n"
					);
				}

			}
			elsif ( $key =~ m/^report_url/ ) {
				if ( lc($val) eq "stdout" ) {
					$Report_url = "stdout";
				}
				elsif ( lc($val) eq "stderr" ) {
					$Report_url = "stderr";
				}
				elsif ( lc($val) eq "null" ) {
					$Report_url = undef;
				}
				else {
					warn
"WARNING : skipped config file $line (line $.) : bad value $val\n";
				}
			}
			else {
				warn
"WARNING : skipped config file $line (line $.) : unknown directive $key\n";
			}
		}
		elsif ( $line =~ m/=/ ) {

			# is an alias definition
			$line =~ s/\s+//g;    # remove all spaces
			my ( $key, $val ) = split ( /=/, $line );
			my $realval = resolv_alias( $., $val, %alias );
			$alias{$key} = $realval;
			debug("alias $key -> $realval\n");
		}

		elsif ( $line =~ m/^!/ ) {

			# begin with ! : it is an exception
			$line =~ s/^!\s*//;

			my @tab     = trait_line($line);
			my $flagdir = pop (@tab);          # and ignore
			my $masq    = pop (@tab);          # and ignore
			foreach my $elem (@tab) {
				if ( test_double($elem) ) {
					warn
"WARNING : skipped config file $line (line $.) : exception $elem already seen in config file\n";
				}
				else {
					debug("exception : $elem\n");
					$Toscan{$elem} = 0;
				}
			}
		}
		else {

			# directory or file to scan
			my @tab     = trait_line($line);
			my $flagdir = pop (@tab);
			my $masqb   = pop (@tab);
			my $masq    = resolv_alias( $., $masqb, %alias );

			foreach my $elem (@tab) {
				if ( test_double($elem) ) {
					warn
"WARNING : skipped config file $line (line $.) : $elem already in config file\n";
				}
				elsif ( -e $elem ) {
					push ( @liste_toscan, $elem );
					debug("toscan : $elem masq $masq\n");
					$Toscan{$elem} = $masq;
				}
				elsif ($Warn_missing_file) {
					warn
"WARNING : skipped  config file $line (line $.) : $elem does not exists\n";
				}
			}    # foreach my $elem
		}
	}    # foreach my $line
	close(INIT);

	# set default values if not founds
	$Verbose            = 0 if ( !defined $Verbose );
	$Warn_dead_symlinks = 0 if ( !defined $Warn_dead_symlinks );
	$Report_full_newdel = 0 if ( !defined $Report_full_newdel );
	$Warn_missing_file  = 0 if ( !defined $Warn_missing_file );

# add some files for internal check : this program ($0), the graphical interface, the config file
	my $masq = resolv_alias( $., "all", %alias );
	my $afick = $0;
	$Toscan{$afick} = $masq;
	my ( $name, $path, undef ) = fileparse($afick);
	my $aficktk = $path . "afick-tk.pl";
	$Toscan{$aficktk} = $masq;
	$Toscan{$config}  = $masq;
	push ( @liste_toscan, $afick, $aficktk, $config );

	debug("-------- end of config file -------------\n");

	return @liste_toscan;
}
#######################################################
# display
# convert fields in a human way
sub display($@) {
	my $elem = shift (@_);
	my @info = @_;

	if ( $elem == 1 ) {

		# device is in hexa, ex 309 == major 3, minor 9 == /dev/hda9
		return sprintf "%lx", $info[$elem];
	}
	elsif ( $elem == 3 ) {

		# perm is in octal
		return sprintf "%lo", $info[$elem];
	}
	elsif ( $elem >= 9 ) {

		# dates
		if ( $info[$elem] ) {
			return gmtime( $info[$elem] );
		}
		else {
			return 0;
		}
	}
	else {
		return $info[$elem];
	}
}

#######################################################
# display_all
# convert in a human way all the record
sub display_all($) {
	my $record = shift (@_);
	my @fileinfo = split ( /\|/, $record );

	my $output;

	my $max = scalar @fileinfo;
	for ( my $i = 0 ; $i < $max ; $i++ ) {
		$output .= display( $i, @fileinfo ) . "|";
	}
	return $output;
}
#######################################################
# md5sum :
# build a MD5 checksum on the given file
sub md5sum($) {
	my $name = shift (@_);    # file name

	my $ctx = Digest::MD5->new;

	if ( open( FILE, "< $name" ) ) {
		binmode(FILE);
		$ctx->addfile(*FILE);
		close(FILE);
		my $sum = $ctx->b64digest;
		return $sum;
	}
	else {
		warn "WARNING : can not open $name for checksum: $!\n";
		return 0;
	}
}
#######################################################
# file_info :
# build a complete info on the given file
sub file_info($$) {
	my $name = shift (@_);
	my $masq = shift (@_);

	debug("$name|");
	my $sum;
	if ( $masq !~ m/h/ ) {
		$sum = 0;
	}
	elsif ( -f $name ) {

		$sum = md5sum($name);
	}
	else {
		$sum = 0;
	}

	my @fileinfo = lstat($name);

	# Each file gets a pipe-delimited entry, with format as follows:
	#
	# field 0 = MD5 hash
	# field 1 = device number of filesystem
	# field 2 = inode number
	# field 3 = file mode (type/permissions)
	# field 4 = number of hard links to file
	# field 5 = uid of the file
	# field 6 = gid of the file
	# field 7 = total size of file, in bytes
	# field 8 = actual number of blocks allocated
	# field 9 = last access time in seconds since the epoch
	# field 10 = last modify time (since epoch)
	# field 11 =  inode change time
	#
	# See the perl "stat" command or man stat() for more information.

	# default value is 0
	my (
		$device, $inode, $perm,  $link,  $uid, $gid,
		$size,   $bloc,  $atime, $mtime, $ctime
	  )
	  = ( 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 );
	$device = $fileinfo[0]  if ( $masq =~ m/d/ );
	$inode  = $fileinfo[1]  if ( $masq =~ m/i/ );
	$perm   = $fileinfo[2]  if ( $masq =~ m/p/ );
	$link   = $fileinfo[3]  if ( $masq =~ m/n/ );
	$uid    = $fileinfo[4]  if ( $masq =~ m/u/ );
	$gid    = $fileinfo[5]  if ( $masq =~ m/g/ );
	$size   = $fileinfo[7]  if ( $masq =~ m/s/ );
	$atime  = $fileinfo[8]  if ( $masq =~ m/a/ );
	$mtime  = $fileinfo[9]  if ( $masq =~ m/m/ );
	$ctime  = $fileinfo[10] if ( $masq =~ m/c/ );
	$bloc   = $fileinfo[12] if ( $masq =~ m/b/ );

	my $entry =
"$sum|$device|$inode|$perm|$link|$uid|$gid|$size|$bloc|$atime|$mtime|$ctime";

	debug( display_all($entry) . "\n" );

	return $entry;
}
#######################################################
# compare_entry :
# compare 2 file infos and print the difference
sub compare_entry(@@) {
	my $entry1 = shift (@_);
	my $entry2 = shift (@_);

	my @fileinfo1 = split ( /\|/, $entry1 );
	my @fileinfo2 = split ( /\|/, $entry2 );

	my @field = (
		'MD5       ',
		'device    ',
		'inode     ',
		'file mode ',
		'links     ',
		'uid       ',
		'gid       ',
		'filesize  ',
		'blocs     ',
		'atime     ',
		'mtime     ',
		'ctime     '
	);

	# Each file gets a pipe-delimited entry, with format as follows:
	my $i;
	my $max = scalar @fileinfo1;
	for ( $i = 0 ; $i < $max ; $i++ ) {
		if ( $fileinfo1[$i] ne $fileinfo2[$i] ) {
			report( "\t$field[$i]\t : "
				  . display( $i, @fileinfo1 ) . "\t"
				  . display( $i, @fileinfo2 )
				  . "\n" );
		}    # if fileinfo ne
	}    # for
}

#######################################################
# is_exception :
# return true if the file name belong exception list
sub is_exception($) {
	my $elem = shift (@_);
	return ( exists( $Toscan{$elem} ) && ( $Toscan{$elem} eq "0" ) );
}
#######################################################
# parcours :
# the general sub to scan a directory (recursive)
# should be called by config list entries
sub parcours($$$$) {
	my $rep     = shift (@_);    # file or directory to scan
	my $traite  = shift (@_);    # sub to call
	my $masq    = shift (@_);    # herited masq
	my $newflag = shift (@_);    # herited new flag

	#debug ("parcours : $rep $masq\n");

	my @liste;

	# first build a list of file
	if ( -d $rep ) {
		return if ( $rep =~ /\/\.\.?$/ );

		if ( exists $Toscan{$rep} ) {
			$masq = $Toscan{$rep};

			debug("change masq to $masq for $rep\n");
		}
		$traite->( $rep, $masq, $newflag );

		# equal : do not scan into the directory
		return if ( exists $Onlythis{$rep} );

		if ( !-l $rep ) {

			if ( !opendir( DIR, $rep ) ) {
				warn "WARNING : can not open directory $rep : $!\n";
				return;
			}
			my $elem;
			while ( defined( $elem = readdir(DIR) ) ) {
				next if $elem =~ /^\.\.?$/;

				#debug( "readdir : $elem");

				# equal : do not scan sub-dir
				next
				  if (  ( exists $Onlydir{$rep} )
					and ( -d $rep . "/" . $elem ) );

				# special case : / to avoid //
				if ( $rep eq "/" ) {
					push ( @liste, '/' . $elem );
				}
				else {
					push ( @liste, $rep . '/' . $elem );
				}
			}
			closedir(DIR);
		}
		else {

			# on ne suit pas les liens sur les repertoires
			debug("$rep skipped (link on directory)\n");
		}
	}
	else {
		push ( @liste, $rep );
	}

	parcours2($rep, $traite, $masq, $Newfiles{$rep}, @liste);
}
#######################################################

sub parcours2($$@) {
	my $rep     = shift (@_);    # file or directory to scan
	my $traite  = shift (@_);    # sub to call
	my $masq    = shift (@_);    # herited masq
	my $newflag = shift (@_);    # herited new flag
	my @liste   = @_;

	foreach my $fic (@liste) {
		next if $fic =~ /^\.\.?$/;
		next if ( is_exception($fic) );

		# heritage, set to all by default (ex : -l action)
		my $lmasq = $masq || "pdinugsbmch";

		# surcharge
		if ( exists $Toscan{$fic} ) {
			$lmasq = $Toscan{$fic};

			#debug ("change masq to $lmasq for $fic");
		}

		debug("parcours : $fic\n");

		if ( -d $fic ) {
			parcours( $fic, $traite, $lmasq, $newflag );
		}    # directory
		elsif ( -l $fic ) {
			my $whatlink = readlink($fic);

			# is the link an absolute path ?
			if ( !( $whatlink =~ m/^\// ) ) {
				$whatlink = $rep . "/" . $whatlink;
			}
			if ( !-e $whatlink ) {

				$Dangling{$fic} = $whatlink;
				next;
			}
			else {
				$traite->( $fic, $lmasq, $Newfiles{$rep} );
			}
		}    # link
		else {
			$traite->( $fic, $lmasq, $Newfiles{$rep} );
		}    # normal file
	}    # foreach
}
#######################################################
# wanted_create
# process a file in create mode
sub wanted_create($$) {
	my $name = shift (@_);
	my $masq = shift (@_);

	$Finded{$name} = 1;

	my $entry = file_info( $name, $masq );

	$Hashfile{$name} = $entry;
}
#######################################################
# end_report
sub end_report() {

	# Generate a MD5 for the dbm file.
	my $sum = md5sum( $database . $database_ext );

	report( "-" x 65 );
	report("\nMD5 hash of $database => $sum\n\n");
	write_control($sum);
}
#######################################################
# create (base liste)
# create a new database (empty the existing one if exists)
sub create($@) {
	my $database = shift (@_);
	my @listerep = @_;
	tie( %Hashfile, $Dbm, $database, O_WRONLY | O_CREAT, 0666 )
	  or die "abort (create) : can not open $Dbm $database : $!";

	# delete all previous keys
	%Hashfile = ();

	foreach my $elem (@listerep) {

		# to skip already scanned files
		next if ( exists $Finded{$elem} );

		parcours( $elem, \&wanted_create, "", "" );
	}
	my $nbscan = scalar( keys %Finded );
	untie %Hashfile;
	print_dangling();

	report("\nHash database created successfully. $nbscan files entered.\n");

	end_report();
}
#######################################################
# wanted_update
# process a file in update mode
sub wanted_update($$) {
	my $name    = shift (@_);
	my $masq    = shift (@_);
	my $newflag = shift (@_);

	$Finded{$name} = 1;

	my $entry = file_info( $name, $masq );

	if ( $Hashfile{$name} ) {

		# test for changes
		my $old_entry = $Hashfile{$name};
		if ( $old_entry ne $entry ) {

			# keep values to compare
			$Oldval{$name} = $old_entry;
			$Newval{$name} = $entry;
			$Nbmod++;
			$Hashfile{$name} = $entry if ($Update);
		}
	}
	else {

		# new entry
		if ( defined $newflag ) {
			$Newfiles{$name} = $newflag + 1;
		}
		else {
			$Newfiles{$name} = 1;
		}

		$Newval{$name} = $entry;

		#debug ("$name heritage : $newflag value $Newfiles{$name}\n");
		$Hashfile{$name} = $entry if ($Update);
	}
}    # end of wanted
#######################################################
# update :
# compare or update a database with the system
sub update($@) {
	my $database = shift (@_);
	my @listerep = @_;
	tie( %Hashfile, $Dbm, $database, O_RDWR, 0666 )
	  or die "abort (update) : can not open $Dbm $database : $!";

	calc_control();

	# scan file list
	foreach my $elem (@listerep) {

		# skip already scanned
		next if ( exists $Finded{$elem} );

		parcours( $elem, \&wanted_update, "", "" );
	}

	# analysis
	my $nbchange = $Nbmod;
	my $nbdelete = 0;
	my $nbnew    = 0;
	if ( !$File ) {

		# look for new files
		my $nb = 0;
		foreach my $key ( sort keys %Newfiles ) {

			# report only if asked all ($Report_full_newdel) or first level
			if ($Report_full_newdel) {
				report( "new " . file_type($key) . ":$key\n" );
			}
			elsif ( $Newfiles{$key} == 1 ) {
				if ($nb) {
					report("  number of new files : $nb\n");
				}
				report( "new " . file_type($key) . ":$key\n" );
				$nb = 0;
			}
			else {
				$nb++;
			}

			$Nbmod++;
			$nbnew++;
		}
		if ($nb) {
			report("  number of new files : $nb\n");
		}

		# look for deleted files
		my %deleted;
		$nb = 0;
		foreach my $key ( sort keys %Hashfile ) {
			if ( !exists( $Finded{$key} ) ) {
				if ($Report_full_newdel) {
					report( "deleted " . file_type($key) . " :$key\n" );
					$nb = 0;
				}
				else {
					my ( $name, $path, undef ) = fileparse($key);
					if ( exists $deleted{$path} ) {
						$deleted{ $key . "/" } = $deleted{$path} + 1;
						$nb++;
					}
					else {
						$deleted{ $key . "/" } = 1;
						if ($nb) {
							report("  number of deleted files : $nb\n");
						}
						$nb = 0;
						report( "deleted " . file_type($key) . " :$key\n" );
					}
				}
				delete $Hashfile{$key} if ($Update);
				$Nbmod++;
				$nbdelete++;
			}
		}
		if ($nb) {
			report("  number of deleted files : $nb\n");
		}
	}
	print_changed(0);
	print_dangling();

	report("\ndetailed changes\n") if $nbchange;
	print_changed(1);

	untie %Hashfile;

	if ($Update) {
		report("\nHash database updated successfully : ");
	}
	else {
		report("\nHash database : ");
	}
	my $nbscan   = scalar( keys %Finded );
	my $dangling = scalar( keys %Dangling );
	report(
"$nbscan files scanned, $Nbmod changed (new : $nbnew; delete : $nbdelete; changed : $nbchange; dangling : $dangling)\n"
	);

	end_report();

	# return a status
	return oct( join ( '', '0b', map { $_ ? 1 : 0;} ($nbnew, $nbdelete, $nbchange, $dangling) ) );
}

#######################################################
# print_dangling
# print all danglink symbolic links
sub print_dangling() {
	return unless ($Warn_dead_symlinks);
	foreach my $key ( keys %Dangling ) {
		report("Dangling link : [$key] => [$Dangling{$key}]\n");
	}
}

#######################################################
# print_dbm
# print the database content
sub print_dbm($) {
	my $database = shift (@_);
	tie( %Hashfile, $Dbm, $database, O_RDONLY, 0666 )
	  or die "abort (print_dbm) : Couldn't tie $Dbm file $database : $!";

	calc_control();

	foreach my $key ( sort keys %Hashfile ) {
		report( file_type($key) . " $key|"
			  . display_all( $Hashfile{$key} )
			  . "\n" );
	}
	report( "number of file : " . scalar( keys %Hashfile ) . "\n" );
	untie %Hashfile;
}
#######################################################
# print a summary or detailed info on changes
sub print_changed($) {
	my $detailed = shift (@_);
	foreach my $key ( sort keys %Oldval ) {
		report( "changed " . file_type($key) . ":$key\n" );
		compare_entry( $Oldval{$key}, $Newval{$key} ) if ($detailed);
	}
}
#######################################################
# print internal flag state
sub print_config() {
	debug("Verbose=$Verbose\n");
	debug("Warn_dead_symlinks=$Warn_dead_symlinks\n");
	debug("Report_full_newdel=$Report_full_newdel\n");
	debug("Warn_missing_file=$Warn_missing_file\n");
	debug("report_url=$Report_url\n");

	#debug("Output_format=$Output_format\n");
}
#######################################################
# usage
# print some help
sub usage($) {
	my $version = shift (@_);
	print "Usage: afick [options]\n\n";
	print "Options:\n\n";
	print "--config_file|-c config\nname of config file to use\n\n";
	print
"--full_newdel|-f,(--nofull_newdel)\n(do not) report full information on new and deleted directories, default is no\n\n";
	print "--help|-h\nthis help\n\n";
	print "--init|-i\nInitialize the hash.dbm database (need -c)\n\n";
	print "--compare|-k\ncompare the hash.dbm database (need -c)\n\n";
	print
	  "--list|-l \"fic1 fic2 ...\"\ncheck the files given in arg (need -c)\n\n";
	print
"--missing_files|-m,(nomissing_files)\n(do not) warn about files declared in config files which does not exists, default is no\n\n";
	print "--print|-p\nprint content of dbm database (need -c)\n\n";
	print
"--dead_symlinks|-s,(--nodead_symlinks)\n(do not) warn about dead symlinks, default is no\n\n";
	print
"--verbose|-v,(--noverbose)\n(not in) verbose mode (for debugging purpose), default is no\n\n";
	print "--version|-V\nprint version\n\n";
	print "--update|-u\ncompare and update the hash.dbm database (need -c)\n\n";
	print "\n";

	print

	  "#################################################################
This script is intended to provide a means for
detecting changes made to files, via a regular
comparison of MD5 hashes to an established baseline. 
#################################################################
Copyright (c) 2002 Eric Gerbier <gerbier\@users.sourceforge.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
#################################################################\n";
}

#######################################################
sub version($) {
	my $version = shift (@_);
	print "\n";
	print "afick : another file integrity checker\nversion $version\n";
}
#######################################################
# read database previous checksum and compare with new one
sub read_control($) {
	my $newchecksum  = shift (@_);
	my $control_file = $database . $control_ext;

	my $oldchecksum;
	if ( open( CONTROL, "< $control_file" ) ) {
		$oldchecksum = <CONTROL>;
		chomp($oldchecksum);
		close(CONTROL);
		if ( $oldchecksum ne $newchecksum ) {
			warn
"WARNING afick internal change : database $database (see below)\n";
		}
	}
	else {
		warn "WARNING can not read control file $control_file : $!\n";
	}
}
#######################################################
# write database checksum for next run
sub write_control ($) {
	my $checksum     = shift (@_);                 # the string to write
	my $control_file = $database . $control_ext;

	# to avoid Tainted warnings
	$control_file = $1 if ( $control_file =~ /(.+)/ );

	if ( open( CONTROL, "> $control_file" ) ) {
		print CONTROL "$checksum";
		close(CONTROL);
	}
	else {
		warn "WARNING can not write in control file $control_file : $!\n";
	}
}
#######################################################
# compare old and new checksum for afick main components
sub control ($) {
	my $name = shift (@_);    # file name to control

	my $newchecksum = md5sum($name);
	my $oldchecksum = get_md5($name);

	if ( !defined $oldchecksum ) {

		# first run
		return;
	}
	elsif ( $oldchecksum ne $newchecksum ) {
		warn "WARNING afick internal change : $name (see below)\n";
	}
}
#######################################################
# check if afick change since last run
sub calc_control() {

	my $afick = $0;
	my ( $name, $path, undef ) = fileparse($afick);
	my $aficktk = $path . "afick-tk.pl";

	control($afick);         # afick.pl
	control($aficktk);       # afick-tk.pl
	control($configfile);    # afick-tk.pl

	my $database_check = md5sum( $database . $database_ext );
	read_control($database_check);

}
########################## main #############################

if ( $^O =~ m/linux/ ) {
	$default_conf = "/etc/afick.conf";
}
elsif ( $^O =~ m/^MSWin/ ) {
	$default_conf = "xp.conf";
}
else {
	$default_conf = "afick.conf";
}

$| = 1;

Getopt::Long::Configure('no_ignore_case');
unless (
	GetOptions(
		"config_file|c=s"  => \$configfile,
		"full_newdel|f!"   => \$Report_full_newdel,
		"help|h"           => \$help,
		"init|i"           => \$init,
		"compare|k"        => \$compare,
		"list|l=s"         => \$list,
		"missing_files|m!" => \$Warn_missing_file,
		"print|p"          => \$print,
		"dead_symlinks|s!" => \$Warn_dead_symlinks,
		"verbose|v!"       => \$Verbose,
		"version|V"        => \$Version,
		"update|u"         => \$Update
	)
  )
{
	usage($version);
	die "abort : incorrect option\n";
}

# -h : help
if ($help) {
	usage($version);
	exit;
}

# -V : version
if ($Version) {
	version($version);
	exit;
}

if ($configfile) {
	debug("config file : $configfile\n");
}
elsif ( -e $default_conf ) {
	$configfile = $default_conf;
}
else {
	usage($version);
	die
"abort : missing configfile name (-c flag) and default config file $default_conf\n";
}
@listerep = read_config($configfile);

# no we should have a database name
if ( !$database ) {
	usage($version);
	die
"abort : missing database name in flag list (-b) and in config file $configfile\n";
}

# actions
if ($list) {

	# -l : liste
	#@listerep = trait_config( split ( ? ?, $liste ) );
	@listerep = split ( ? ?, $list );
	$File = 1;
	exit update( $database, @listerep );
}
elsif ($print) {

	# -p : print
	print_dbm($database);
	exit;
}
elsif ($init) {

	# -i : init
	create( $database, @listerep );
}
elsif ($Update) {

	# -u : update
	exit update( $database, @listerep );
}
elsif ($compare) {

	# -k : check
	exit update( $database, @listerep );
}
else {
	usage($version);
	die "abort : no action to do (-i, -u, -k, -p) \n";
}
#######################################################
__END__

=head1 NAME

afick : Another File Integrity Checker

=head1 SYNOPSIS

=head1 OPTIONS

=head1 DESCRIPTION

=cut
